@@ -51,8 +51,7 @@ typedef void (* NautilusIconCallback) (NautilusIconData *icon_data,
 typedef struct {
 	int x;
 	int y;
-	double scale_x;
-	double scale_y;
+	double scale;
 } NautilusIconPosition;
 
 typedef enum {
@@ -129,6 +128,7 @@ typedef struct {
 						   NautilusIconData *data,
 						   GList **emblem_icons,
 						   char **embedded_text,
+						   gboolean need_large_embeddded_text,
 						   gboolean *embedded_text_needs_loading,
 						   gboolean *has_window_open);
 	void         (* get_icon_text)            (NautilusIconContainer *container,
@@ -145,7 +145,8 @@ typedef struct {
 						   NautilusIconData *icon_b);
 	void         (* start_monitor_top_left)   (NautilusIconContainer *container,
 						   NautilusIconData *data,
-						   gconstpointer client);
+						   gconstpointer client,
+						   gboolean large_text);
 	void         (* stop_monitor_top_left)    (NautilusIconContainer *container,
 						   NautilusIconData *data,
 						   gconstpointer client);
@@ -194,6 +195,7 @@ typedef struct {
         void         (* icon_removed)             (NautilusIconContainer *container,
                                                    NautilusIconData *data);
         void         (* cleared)                  (NautilusIconContainer *container);
+	gboolean     (* start_interactive_search) (NautilusIconContainer *container);
 } NautilusIconContainerClass;
 
 /* GtkObject */
@@ -204,7 +206,8 @@ GtkWidget *       nautilus_icon_container_new                           (void);
 /* adding, removing, and managing icons */
 void              nautilus_icon_container_clear                         (NautilusIconContainer  *view);
 gboolean          nautilus_icon_container_add                           (NautilusIconContainer  *view,
-									 NautilusIconData       *data);
+									 NautilusIconData       *data,
+									 gboolean                has_lazy_position);
 void              nautilus_icon_container_layout_now                    (NautilusIconContainer *container);
 gboolean          nautilus_icon_container_remove                        (NautilusIconContainer  *view,
 									 NautilusIconData       *data);
@@ -285,5 +288,8 @@ void              nautilus_icon_container_set_use_drop_shadows          (Nautilu
 									 gboolean                use_drop_shadows);
 char*             nautilus_icon_container_get_icon_description          (NautilusIconContainer  *container,
                                                                          NautilusIconData       *data);
+gboolean          nautilus_icon_container_get_allow_moves               (NautilusIconContainer  *container);
+void              nautilus_icon_container_set_allow_moves               (NautilusIconContainer  *container,
+									 gboolean                allow_moves);
 
 #endif /* NAUTILUS_ICON_CONTAINER_H */
