@@ -78,6 +78,11 @@ typedef struct {
 	GtkWidget *popup_menu;
 	GtkWidget *popup_menu_remove_item;
 	GtkWidget *popup_menu_rename_item;
+	GtkWidget *popup_menu_separator_item;
+	GtkWidget *popup_menu_mount_item;
+	GtkWidget *popup_menu_unmount_item;
+	GtkWidget *popup_menu_eject_item;
+	GtkWidget *popup_menu_format_item;
 } NautilusPlacesSidebar;
 
 typedef struct {
@@ -96,6 +101,7 @@ enum {
 	PLACES_SIDEBAR_COLUMN_ROW_TYPE,
 	PLACES_SIDEBAR_COLUMN_URI,
 	PLACES_SIDEBAR_COLUMN_DRIVE,
+	PLACES_SIDEBAR_COLUMN_VOLUME,
 	PLACES_SIDEBAR_COLUMN_NAME,
 	PLACES_SIDEBAR_COLUMN_ICON,
 	PLACES_SIDEBAR_COLUMN_INDEX,
@@ -129,9 +135,14 @@ G_DEFINE_TYPE_WITH_CODE (NautilusPlacesSidebarProvider, nautilus_places_sidebar_
 
 
 static GtkTreeIter
-add_place (GtkListStore *store, PlaceType place_type,
-	   const char *name, const char *icon, const char *uri,
-	   GnomeVFSDrive *drive, const int index)
+add_place (GtkListStore *store,
+	   PlaceType place_type,
+	   const char *name,
+	   const char *icon,
+	   const char *uri,
+	   GnomeVFSDrive *drive,
+	   GnomeVFSVolume *volume,
+	   const int index)
 {
 	GdkPixbuf            *pixbuf;
 	GtkTreeIter           iter;
@@ -143,6 +154,7 @@ add_place (GtkListStore *store, PlaceType place_type,
 			    PLACES_SIDEBAR_COLUMN_NAME, name,
 			    PLACES_SIDEBAR_COLUMN_URI, uri,
 			    PLACES_SIDEBAR_COLUMN_DRIVE, drive,
+			    PLACES_SIDEBAR_COLUMN_VOLUME, volume,
 			    PLACES_SIDEBAR_COLUMN_ROW_TYPE, place_type,
 			    PLACES_SIDEBAR_COLUMN_INDEX, index,
 			    -1);
@@ -180,7 +192,8 @@ update_places (NautilusPlacesSidebar *sidebar)
 		mount_uri = gnome_vfs_get_uri_from_local_path (g_get_home_dir ());
 		display_name = g_filename_display_basename (g_get_home_dir ());
 		last_iter = add_place (sidebar->store, PLACES_BUILT_IN,
-				       display_name, "gnome-fs-home", mount_uri, NULL, 0);
+				       display_name, "gnome-fs-home",
+				       mount_uri, NULL, NULL, 0);
 		g_free (display_name);
 		if (strcmp (location, mount_uri) == 0) {
 			gtk_tree_selection_select_iter (selection, &last_iter);
@@ -190,7 +203,8 @@ update_places (NautilusPlacesSidebar *sidebar)
 
 	mount_uri = gnome_vfs_get_uri_from_local_path (desktop_path);
 	last_iter = add_place (sidebar->store, PLACES_BUILT_IN,
-			       _("Desktop"), "gnome-fs-desktop", mount_uri, NULL, 0);
+			       _("Desktop"), "gnome-fs-desktop",
+			       mount_uri, NULL, NULL, 0);
 	if (strcmp (location, mount_uri) == 0) {
 		gtk_tree_selection_select_iter (selection, &last_iter);
 	}	
@@ -199,7 +213,8 @@ update_places (NautilusPlacesSidebar *sidebar)
 	
  	mount_uri = "file:///"; /* No need to strdup */
 	last_iter = add_place (sidebar->store, PLACES_BUILT_IN,
-			       _("File System"), "gnome-dev-harddisk", mount_uri, NULL, 0);
+			       _("File System"), "gnome-dev-harddisk",
+			       mount_uri, NULL, NULL, 0);
 	if (strcmp (location, mount_uri) == 0) {
 		gtk_tree_selection_select_iter (selection, &last_iter);
 	}	
@@ -229,7 +244,8 @@ update_places (NautilusPlacesSidebar *sidebar)
 				mount_uri = gnome_vfs_volume_get_activation_uri (volume);
 				name = gnome_vfs_volume_get_display_name (volume);
 				last_iter = add_place (sidebar->store, PLACES_MOUNTED_VOLUME,
-						       name, icon, mount_uri, drive, 0);
+						       name, icon, mount_uri,
+						       drive, volume, 0);
 				if (strcmp (location, mount_uri) == 0) {
 					gtk_tree_selection_select_iter (selection, &last_iter);
 				}
@@ -246,7 +262,8 @@ update_places (NautilusPlacesSidebar *sidebar)
 			icon = gnome_vfs_drive_get_icon (drive);
 			name = gnome_vfs_drive_get_display_name (drive);
 			last_iter = add_place (sidebar->store, PLACES_BUILT_IN,
-					       name, icon, NULL, drive, 0);
+					       name, icon, NULL,
+					       drive, NULL, 0);
 			g_free (icon);
 			g_free (name);
 		}		
@@ -270,7 +287,8 @@ update_places (NautilusPlacesSidebar *sidebar)
 		mount_uri = gnome_vfs_volume_get_activation_uri (volume);
 		name = gnome_vfs_volume_get_display_name (volume);
 		last_iter = add_place (sidebar->store, PLACES_MOUNTED_VOLUME,
-				       name, icon, mount_uri, NULL, 0);
+				       name, icon, mount_uri,
+				       NULL, volume, 0);
 		if (strcmp (location, mount_uri) == 0) {
 			gtk_tree_selection_select_iter (selection, &last_iter);
 		}
@@ -302,7 +320,8 @@ update_places (NautilusPlacesSidebar *sidebar)
 		icon = nautilus_bookmark_get_icon (bookmark);
 		mount_uri = nautilus_bookmark_get_uri (bookmark);
 		last_iter = add_place (sidebar->store, PLACES_BOOKMARK,
-				       name, icon, mount_uri, NULL, index);
+				       name, icon, mount_uri,
+				       NULL, NULL, index);
 		if (strcmp (location, mount_uri) == 0) {
 			gtk_tree_selection_select_iter (selection, &last_iter);
 		}
@@ -314,6 +333,13 @@ update_places (NautilusPlacesSidebar *sidebar)
 	
 }
 
+static gboolean
+update_places_cb (gpointer data)
+{
+	update_places (NAUTILUS_PLACES_SIDEBAR (data));
+	return FALSE;
+}
+
 static gboolean
 shortcuts_row_separator_func (GtkTreeModel *model,
 			      GtkTreeIter  *iter,
@@ -343,20 +369,40 @@ static void
 volume_unmounted_callback (GnomeVFSVolumeMonitor *volume_monitor,
 			   GnomeVFSVolume *volume,
 			   NautilusPlacesSidebar *sidebar)
+{
+	/* At this point the volume still appears to be mounted.
+	 * GnomeVFS will update its list after the signal finished
+	 * to emit, so we delay the update.
+	 */
+	g_idle_add (update_places_cb, sidebar);
+}
+
+static void
+drive_disconnected_callback (GnomeVFSVolumeMonitor *volume_monitor,
+			     GnomeVFSDrive         *drive,
+			     NautilusPlacesSidebar *sidebar)
 {
 	update_places (sidebar);
 }
 
 static void
-mount_volume_callback (gboolean succeeded,
-		       char *error,
-		       char *detailed_error,
-		       gpointer user_data)
+drive_connected_callback (GnomeVFSVolumeMonitor *volume_monitor,
+			  GnomeVFSDrive         *drive,
+			  NautilusPlacesSidebar *sidebar)
+{
+	update_places (sidebar);
+}
+
+static void
+volume_op_callback (gboolean succeeded,
+		    char *error,
+		    char *detailed_error,
+		    gpointer user_data)
 {
 	if (!succeeded) {
 		if (*error == 0 &&
 		    detailed_error != NULL && *detailed_error == 0) {
-			/* This means the mount command displays its own errors */
+			/* This means the command displays its own errors */
 			return;
 		}
 		eel_show_error_dialog_with_details (error, NULL,
@@ -396,7 +442,7 @@ row_activated_callback (GtkTreeView *tree_view,
 		gtk_tree_model_get 
 			(model, &iter, PLACES_SIDEBAR_COLUMN_DRIVE, &drive, -1);
 		if (drive != NULL) {
-			gnome_vfs_drive_mount (drive, mount_volume_callback, sidebar);
+			gnome_vfs_drive_mount (drive, volume_op_callback, sidebar);
 			gnome_vfs_drive_unref (drive);
 		}
 	}
@@ -887,30 +933,104 @@ bookmarks_popup_menu_detach_cb (GtkWidget *attach_widget,
 	sidebar->popup_menu = NULL;
 	sidebar->popup_menu_remove_item = NULL;
 	sidebar->popup_menu_rename_item = NULL;
+	sidebar->popup_menu_separator_item = NULL;
+	sidebar->popup_menu_mount_item = NULL;
+	sidebar->popup_menu_unmount_item = NULL;
+	sidebar->popup_menu_eject_item = NULL;
+	sidebar->popup_menu_format_item = NULL;
+}
+
+static gboolean
+eject_for_type (GnomeVFSDeviceType type)
+{
+	switch (type) {
+	case GNOME_VFS_DEVICE_TYPE_CDROM:
+	case GNOME_VFS_DEVICE_TYPE_ZIP:
+	case GNOME_VFS_DEVICE_TYPE_JAZ:
+		return TRUE;
+	default:
+		return FALSE;
+	}
+}
+
+static void
+check_visibility (GnomeVFSVolume *volume,
+		  GnomeVFSDrive  *drive,
+		  gboolean       *show_mount,
+		  gboolean       *show_unmount,
+		  gboolean       *show_eject,
+		  gboolean       *show_format)
+{
+	*show_mount = FALSE;
+	*show_unmount = FALSE;
+	*show_eject = FALSE;
+	*show_format = FALSE;
+
+	if (volume != NULL) {
+		*show_unmount = TRUE;
+		*show_eject = eject_for_type (gnome_vfs_volume_get_device_type (volume));
+	} else if (drive != NULL) {
+		*show_eject = eject_for_type (gnome_vfs_drive_get_device_type (drive));
+		if (gnome_vfs_drive_is_mounted (drive)) {
+			*show_unmount = TRUE;
+		} else {
+			*show_mount = TRUE;
+		}
+
+		if (gnome_vfs_drive_get_device_type (drive) == GNOME_VFS_DEVICE_TYPE_FLOPPY &&
+		    g_find_program_in_path ("gfloppy")) {
+			*show_format = TRUE;
+		}
+	}
 }
 
 static void
 bookmarks_check_popup_sensitivity (NautilusPlacesSidebar *sidebar)
 {
-  GtkTreeIter iter;
-  PlaceType type; 
-  GtkTreeSelection *selection;
+	GtkTreeIter iter;
+	PlaceType type; 
+	GtkTreeSelection *selection;
+	GnomeVFSDrive *drive = NULL;
+	GnomeVFSVolume *volume = NULL;
+	gboolean show_mount;
+	gboolean show_unmount;
+	gboolean show_eject;
+	gboolean show_format;
+	
+	type = PLACES_BUILT_IN;
 
-  type = PLACES_BUILT_IN;
+	if (sidebar->popup_menu == NULL) {
+		return;
+	}
 
-  if (sidebar->popup_menu == NULL)
-    return;
+	selection = gtk_tree_view_get_selection (sidebar->tree_view);
 
-  selection = gtk_tree_view_get_selection (sidebar->tree_view);
+	if (gtk_tree_selection_get_selected (selection, NULL, &iter)) {
+		gtk_tree_model_get (GTK_TREE_MODEL (sidebar->store), &iter,
+				    PLACES_SIDEBAR_COLUMN_ROW_TYPE, &type,
+				    PLACES_SIDEBAR_COLUMN_DRIVE, &drive,
+ 				    PLACES_SIDEBAR_COLUMN_VOLUME, &volume,
+				    -1);
+	}
 
-  if (gtk_tree_selection_get_selected (selection, NULL, &iter)) {
- 	  gtk_tree_model_get (GTK_TREE_MODEL (sidebar->store), &iter,
-			      PLACES_SIDEBAR_COLUMN_ROW_TYPE, &type,
-			      -1);
-  }
+	gtk_widget_set_sensitive (sidebar->popup_menu_remove_item, (type == PLACES_BOOKMARK));
+	gtk_widget_set_sensitive (sidebar->popup_menu_rename_item, (type == PLACES_BOOKMARK));
 
-  gtk_widget_set_sensitive (sidebar->popup_menu_remove_item, (type == PLACES_BOOKMARK));
-  gtk_widget_set_sensitive (sidebar->popup_menu_rename_item, (type == PLACES_BOOKMARK));
+ 	check_visibility (volume, drive,
+ 			  &show_mount, &show_unmount, &show_eject, &show_format);	
+
+	/* We don't want both eject and unmount, since eject
+ 	   unmounts too */
+ 	if (show_eject) {
+ 		show_unmount = FALSE;
+  	}
+	
+	eel_gtk_widget_set_shown (sidebar->popup_menu_separator_item, 
+			show_mount || show_unmount || show_eject || show_format);
+	eel_gtk_widget_set_shown (sidebar->popup_menu_mount_item, show_mount);
+	eel_gtk_widget_set_shown (sidebar->popup_menu_unmount_item, show_unmount);
+	eel_gtk_widget_set_shown (sidebar->popup_menu_eject_item, show_eject);
+	eel_gtk_widget_set_shown (sidebar->popup_menu_format_item, show_format);
 }
 
 /* Callback used when the selection in the shortcuts tree changes */
@@ -921,6 +1041,14 @@ bookmarks_selection_changed_cb (GtkTreeSelection      *selection,
 	bookmarks_check_popup_sensitivity (sidebar);
 }
 
+static gboolean
+get_selected_iter (NautilusPlacesSidebar *sidebar,
+		   GtkTreeIter *iter)
+{
+	GtkTreeSelection *selection;
+	selection = gtk_tree_view_get_selection (sidebar->tree_view);
+	return gtk_tree_selection_get_selected (selection, NULL, iter);
+}
 
 /* Rename the selected bookmark */
 static void
@@ -931,11 +1059,8 @@ rename_selected_bookmark (NautilusPlacesSidebar *sidebar)
 	GtkTreeViewColumn *column;
 	GtkCellRenderer *cell;
 	GList *renderers;
-	GtkTreeSelection *selection;
 	
-	selection = gtk_tree_view_get_selection (sidebar->tree_view);
-	
-	if (gtk_tree_selection_get_selected (selection, NULL, &iter)) {
+	if (get_selected_iter (sidebar, &iter)) {
 		path = gtk_tree_model_get_path (GTK_TREE_MODEL (sidebar->store), &iter);
 		column = gtk_tree_view_get_column (GTK_TREE_VIEW (sidebar->tree_view), 0);
 		renderers = gtk_tree_view_column_get_cell_renderers (column);
@@ -960,13 +1085,10 @@ static void
 remove_selected_bookmarks (NautilusPlacesSidebar *sidebar)
 {
 	GtkTreeIter iter;
-	GtkTreeSelection *selection;
 	PlaceType type; 
 	int index;
 
-	selection = gtk_tree_view_get_selection (sidebar->tree_view);
-	
-	if (!gtk_tree_selection_get_selected (selection, NULL, &iter)) {
+	if (!get_selected_iter (sidebar, &iter)) {
 		return;
 	}
 	
@@ -992,6 +1114,85 @@ remove_shortcut_cb (GtkMenuItem           *item,
 	remove_selected_bookmarks (sidebar);
 }
 
+static void
+mount_shortcut_cb (GtkMenuItem           *item,
+		   NautilusPlacesSidebar *sidebar)
+{
+	GtkTreeIter iter;
+	GnomeVFSDrive *drive;
+
+	if (!get_selected_iter (sidebar, &iter)) {
+		return;
+	}
+
+	gtk_tree_model_get (GTK_TREE_MODEL (sidebar->store), &iter,
+			    PLACES_SIDEBAR_COLUMN_DRIVE, &drive,
+			    -1);
+
+	if (drive != NULL) {
+		gnome_vfs_drive_mount (drive, volume_op_callback, sidebar);
+		gnome_vfs_drive_unref (drive);
+	}
+}
+
+static void
+unmount_shortcut_cb (GtkMenuItem           *item,
+		     NautilusPlacesSidebar *sidebar)
+{
+	GtkTreeIter iter;
+	GnomeVFSVolume *volume;
+	GnomeVFSDrive  *drive;
+
+	if (!get_selected_iter (sidebar, &iter)) {
+		return;
+	}
+
+	gtk_tree_model_get (GTK_TREE_MODEL (sidebar->store), &iter,
+			    PLACES_SIDEBAR_COLUMN_VOLUME, &volume,
+			    PLACES_SIDEBAR_COLUMN_DRIVE, &drive,
+			    -1);
+
+	if (volume != NULL) {
+		gnome_vfs_volume_unmount (volume, volume_op_callback, sidebar);
+	} else if (drive != NULL) {
+		gnome_vfs_drive_unmount (drive, volume_op_callback, sidebar);
+	}
+	gnome_vfs_volume_unref (volume);
+	gnome_vfs_drive_unref (drive);
+}
+
+static void
+eject_shortcut_cb (GtkMenuItem           *item,
+		   NautilusPlacesSidebar *sidebar)
+{
+	GtkTreeIter iter;
+	GnomeVFSVolume *volume;
+	GnomeVFSDrive  *drive;
+
+	if (!get_selected_iter (sidebar, &iter)) {
+		return;
+	}
+
+	gtk_tree_model_get (GTK_TREE_MODEL (sidebar->store), &iter,
+			    PLACES_SIDEBAR_COLUMN_VOLUME, &volume,
+			    PLACES_SIDEBAR_COLUMN_DRIVE, &drive,
+			    -1);
+
+	if (volume != NULL) {
+		gnome_vfs_volume_eject (volume, volume_op_callback, sidebar);
+	} else if (drive != NULL) {
+		gnome_vfs_drive_eject (drive, volume_op_callback, sidebar);
+	}
+	gnome_vfs_volume_unref (volume);
+	gnome_vfs_drive_unref (drive);
+}
+
+static void
+format_shortcut_cb (GtkMenuItem           *item,
+		    NautilusPlacesSidebar *sidebar)
+{
+	g_spawn_command_line_async ("gfloppy", NULL);
+}
 
 /* Handler for GtkWidget::key-press-event on the shortcuts list */
 static gboolean
@@ -1051,6 +1252,39 @@ bookmarks_build_popup_menu (NautilusPlacesSidebar *sidebar)
 	gtk_widget_show (item);
 	gtk_menu_shell_append (GTK_MENU_SHELL (sidebar->popup_menu), item);
 	
+	/* Mount/Unmount/Eject menu items */
+
+	sidebar->popup_menu_separator_item =
+		GTK_WIDGET (eel_gtk_menu_append_separator (GTK_MENU (sidebar->popup_menu)));
+
+	item = gtk_menu_item_new_with_mnemonic (_("_Mount"));
+	sidebar->popup_menu_mount_item = item;
+	g_signal_connect (item, "activate",
+		    G_CALLBACK (mount_shortcut_cb), sidebar);
+	gtk_widget_show (item);
+	gtk_menu_shell_append (GTK_MENU_SHELL (sidebar->popup_menu), item);
+
+	item = gtk_menu_item_new_with_mnemonic (_("_Unmount"));
+	sidebar->popup_menu_unmount_item = item;
+	g_signal_connect (item, "activate",
+		    G_CALLBACK (unmount_shortcut_cb), sidebar);
+	gtk_widget_show (item);
+	gtk_menu_shell_append (GTK_MENU_SHELL (sidebar->popup_menu), item);
+
+	item = gtk_menu_item_new_with_mnemonic (_("_Eject"));
+	sidebar->popup_menu_eject_item = item;
+	g_signal_connect (item, "activate",
+		    G_CALLBACK (eject_shortcut_cb), sidebar);
+	gtk_widget_show (item);
+	gtk_menu_shell_append (GTK_MENU_SHELL (sidebar->popup_menu), item);
+
+	item = gtk_menu_item_new_with_mnemonic (_("_Format"));
+	sidebar->popup_menu_format_item = item;
+	g_signal_connect (item, "activate",
+		    G_CALLBACK (format_shortcut_cb), sidebar);
+	gtk_widget_show (item);
+	gtk_menu_shell_append (GTK_MENU_SHELL (sidebar->popup_menu), item);
+
 	bookmarks_check_popup_sensitivity (sidebar);
 }
 
@@ -1178,6 +1412,7 @@ nautilus_places_sidebar_init (NautilusPlacesSidebar *sidebar)
 					     G_TYPE_INT, 
 					     G_TYPE_STRING,
 					     GNOME_VFS_TYPE_DRIVE,
+					     GNOME_VFS_TYPE_VOLUME,
 					     G_TYPE_STRING,
 					     GDK_TYPE_PIXBUF,
 					     G_TYPE_INT
@@ -1325,6 +1560,10 @@ nautilus_places_sidebar_set_parent_window (NautilusPlacesSidebar *sidebar,
 				 G_CALLBACK (volume_mounted_callback), sidebar, 0);
 	g_signal_connect_object (volume_monitor, "volume_unmounted",
 				 G_CALLBACK (volume_unmounted_callback), sidebar, 0);
+	g_signal_connect_object (volume_monitor, "drive_disconnected",
+				 G_CALLBACK (drive_disconnected_callback), sidebar, 0);
+	g_signal_connect_object (volume_monitor, "drive_connected",
+				 G_CALLBACK (drive_connected_callback), sidebar, 0);
 
 	update_places (sidebar);
 }
